# Hello, this program is written in Python - http://python.org
import os, random, urllib, re, sys, time, threading, Queue
import token
# For more information, see http://sebsauvage.net/python/gossyp/

debug = False   # Set to True to see this fetcher working.

class rfc822parser:
    ''' This class parses RFC822 messages (mail/usenet format), and extracts sentences.
        The messages are cleaned up as much a possible (headers, quoted text, signatures...).
        Example: print '\n\n'.join( rfc822parser(mymessage).getSentences() )       
        
        WARNING: According to the Google license, you are not allowed to make automated
        queries on their search engine. Google will refuse to answer with
        Python/URLLib as User-Agent.
        (I circumvent that with a local proxy (Proxomitron) which re-writes
        User-Agent HTTP header.)
        
        To do: write a custom URL opener which uses a different User-Agent.
    '''
    _sentences_regexp = re.compile('(.{2,}?[\.\?\!]) ',re.DOTALL)

    def __init__(self, message):
        ''' Input: message is the full RFC822 message. '''
        self._sentences = None
        lines = message.splitlines()
        lines = [line.rstrip() for line in lines[:]] # Strip right spaces in each line.

        # Strip all lines until we find an empty line (this is RFC822 format).
        while len(lines[0]) > 0: lines.pop(0)
        # Also strip these empty line before message body (The real message body start just after the empty lines).
        while len(lines[0]) == 0: lines.pop(0)


        lines = [i for i in lines if not i.lstrip().startswith('=')] # Remove all lines starting with '='
        lines = [i for i in lines if not i.lstrip().startswith('-')] # Remove all lines starting with '-'

        # Now remove all lines starting with '>' (they're only quoting a previous message.)
        lines = [i for i in lines if not i.lstrip().startswith('>')]
        lines = [i for i in lines if not i.lstrip().startswith('<')]

        # Try to remove the source of quoting (and cope with the different quoting styles <grin>)
        lines = [i for i in lines if not i.rstrip().endswith(' wrote:')]
        lines = [i for i in lines if not (i.lstrip().startswith('"') and (i.find('wrote in message')>-1))]
        lines = [i for i in lines if not (i.lstrip().startswith('On ') \
                                          and (i[3:6] in ('Mon','Tue','Wed','Thu','Fri','Sat','Sun')))]
        lines = [i for i in lines if not i.lstrip().startswith('In article <')]
        lines = [i for i in lines if not i.lstrip().startswith('From:')]

        # Remove lines containing URLs.
        lines = [i for i in lines if not i.lstrip().startswith('http:')]
        lines = [i for i in lines if not i.lstrip().startswith('ftp:')]
        lines = [i for i in lines if not i.lstrip().startswith('news:')]

        # Remove all lines after the first line containing only '--' (they are signatures)
        try:    lines = lines[:lines.index('--')]
        except: pass

        # Now remove all newlines (CR and LF)
        ctext = ' '.join(lines).strip()

        # Make sure text is properly terminated (for the regexp below).
        if ctext[-1:] not in ['.','!','?']: ctext += '.'
        ctext += ' '

        # remove spaces that are useless.
        while (ctext.find('  ')>-1): ctext = ctext.replace('  ',' ')

        ctext = ctext.replace('...','') # Keep '...' safe.

        # Split into phrases using . ! or ? as separator.
        # The 'try' is used to protected against 'maximum recursion limit exceeded' error.
        # (rare, but can happen on some messages.)
        try:
            split1 = self._sentences_regexp.findall(ctext)
        except:
            return

        # Put back the '...'.
        # Remove sentences that are too short or too long.
        sentences = [i.replace('','...').strip() for i in split1 if (len(i)>5 and len(i)<140)]
        self._sentences = sentences

    def getSentences(self):
        ''' Returns all the sentences in this message.
            Output: a list of string.  Each string contains a sentence. '''
        return self._sentences

class gossypFetcher_googleGroups(threading.Thread):
    ''' Fetches sentences from different Usenet groups using Google Group search engine.
        Returns the most recent message from hundreds of different newsgroups.
        Protocol used: HTTP
        Server used: groups.google.com
    '''
    '''
    Technical details:
    The following URL returns links to 10 most recent messages from talk.* groups:
    http://groups.google.com/groups?as_ugroup=talk.*&as_scoring=d&lr=lang_fr&start=10
    Parameters:
    as_ugroup    Search in this Usenet branch.
    as_scoring   Sort by date (most recent first)
    lr           Get only messages in this language (lang_fr for French, lang_en for English, etc.)
    start        Get page x of results.

    Once fetched, follow link to each message.
    They are URLs containing selm=...

    Operation
    1. HTTP request to query google groups search engine (parameters: group, language, start page)
    2. Parse result and extract individual messages URL.
    3. HTTP request to get each message.
    4. Parse each message to extract sentences.
    '''

    # List of available languages for this fetcher.
    _languages = ['en','fr']

    # List of branches of Usenet we will fetch sentences from:
    # (I chose not to include comp.* to lesser the number of computerish jargon.)
    _groups = ['alt.*','humanities.*','misc.*','news.*','rec.*','sci.*','soc.*','talk.*']

    # Fetch a sentence in the x most recent messages in this Usenet branch.
    _max_start = 500 # = 50 pages.

    # Regular expression to extract individual messages URLs in a Google Groups search result page.
    _messageURL = re.compile('<a href=([^\s]*?selm=[^\s]*?)\s?>')

    def __init__(self, outQueue, inQueue=None):
        ''' inQueue is the command queue (put token in this queue to send orders to this class.)
            outQueue is a Queue object where sentences will be written to. '''
        threading.Thread.__init__(self)
        self.outQueue = outQueue   # Output queue (sentences)
        if inQueue is None: self.inQueue = Queue.Queue()
        else:               self.inQueue = inQueue     # Input queue (tokens)
        self.done = True

    def run(self):
        ''' Start listening for commands (tokens) from the inQueue. '''
        if debug: print 'gossypFetcher_googleGroups: running.'
        while True:
            try:          token = self.inQueue.get()  # Get orders
            except Empty: time.sleep(1)
            if token.shutdown:
                if debug: print 'gossypFetcher_googleGroups: shutting down.'
                break
            elif token.fetch:
                self.fetchSentences()
            else:
                pass

    def fetchSentences(self):
        ''' Fetches sentences from Usenet using Google Groups search engine.
            Writes sentences in the output Queue (self.outQueue). '''
        self.done = False
        if debug: print 'gossypFetcher_googleGroups: fetching.'
        language='en'
        group = random.choice(self._groups)  # Choose a random Usenet branch.
        start = random.randrange(0,self._max_start)  # Choose a random result page.
        if language not in self._languages: language = self._languages[0]

        try:
            # Make a search on Google Groups:
            result_page = urllib.urlopen('http://groups.google.com/groups?as_ugroup=%s&as_scoring=d&lr=%s&start=%d' % (group,language,start) ).read()

            # Extract the messages URLs from this result page.
            urls = []
            for match in self._messageURL.findall(result_page):
                urls.append('http://groups.google.com'+match+'&output=gplain')  # output=gplain for plaintext message (easier to parse).
            
            # Fetch each message and extract sentences from these messages:
            sentences_count = 0
            for url in urls:
                try:
                    sentences = rfc822parser(urllib.urlopen(url).read()).getSentences()
                    sentences_count += len(sentences)
                    for sentence in sentences:
                        self.outQueue.put(sentence)
                except:
                    pass  # Skip this message if there's a problem.
            if debug and sentences_count>0: print 'gossypFetcher_googleGroups: I gave %d sentences.' % sentences_count
        except:
            if debug: print 'gossypFetcher_googleGroups: Problem fetching. Aborting fetching.'
            self.done = True
            return  # Stop fetching if there's a problem.
        if debug: print 'gossypFetcher_googleGroups: Done fetching.'
        self.done = True

if __name__ == "__main__":
    #os.environ['http_proxy'] = 'http://127.0.0.1:3131'
    debug = True
    out = Queue.Queue()
    f = gossypFetcher_googleGroups(out)
    f.start()
    f.inQueue.put(token.Token(fetch=1))
    f.inQueue.put(token.Token(fetch=1))
    f.inQueue.put(token.Token(fetch=1))
    while 1:
        try   :sentence = out.get()   
        except: pass
        print sentence+'\n'
        time.sleep(3)
