# Hello, this program is written in Python - http://python.org
import os, random, urllib, re, sys, time, threading, Queue, htmlentitydefs
#import sha  # for debugging.
import token

# For more information, see http://sebsauvage.net/python/gossyp/

debug = False     # Set to True to see this fetcher working.

class gossypFetcher_YahooMessageBoards(threading.Thread):
    ''' Fetches sentences from different Yahoo message boards.
        Protocol used: HTTP
        Server used: messages.yahoo.com
    '''
    '''
    Technical details:

    This one is a bit more complex than GoogleGroups.
    Yahoo! Message Boards presents a hierarchy of directories.
    Each directory contains one or more subdirectories and/or topics.
    Each topic contains several discussions.
    Each discussion contains several messages.

    Homepage is http://messages.yahoo.com/
    Each message board is readable without login.

    http://messages.yahoo.com/yahoo/index.html
    This is the home page. Contains directories.

    http://messages.yahoo.com/yahoo/...
    or http://news.messages.yahoo.com/yahoo/...
    This is a directory. Contains other directories and topics.

    http://messages.yahoo.com/bbs?...action=topics...
    This is a topic. Contains the most recent discussions on this topic.
    (luckly, the topic page presents the discussions which contain the most recent contributions.)

    http://messages.yahoo.com/bbs?...action=m...mid=1
    This is a discussion. Contains the list of most recent message in this discussion.
    (luckly, the discussion page (which contains the 1st message of the discussions)
     also contains the most recent contributions.)

    http://messages.yahoo.com/bbs?...action=m...mid=...
    This is a message.

    We have to explore this hierarchy.
    That's why this fetcher takes some time for warming up
    before giving its first sentences.
    '''

    # Regular expression for link extraction.
    _href_re = re.compile('<a href="(.*?)">', re.DOTALL + re.IGNORECASE)

    # Regular expression to remove HTML tags.
    _tagsRemove_re = re.compile('<\\?.+?>')

    # Regular expression to replace HTML entities with their ascii code
    _htmlentity_re = re.compile('&(\w+);',re.IGNORECASE)

    # Regular expression to cut a plain text into individual sentences.
    _sentences_regexp = re.compile('(.{2,}?[\.\?\!]) ',re.DOTALL)


    def __init__(self, outQueue, inQueue=None):
        ''' inQueue is the command queue (put token in this queue to send orders to this class.
            outQueue is a Queue object where sentences will be written to. '''
        threading.Thread.__init__(self)
        self._dir_2do   = {}   # Directories to explore.
        self._dir       = {}   # Directories explored.
        self._topic_2do = {}   # Topics to explore.
        self._topic     = {}   # Topics explored.
        self._disc_2do  = {}   # Discussions to explore.
        self._disc      = {}   # Discussions explored.
        self._message   = {}   # Individual messages
        # (We do not store explored messages.)
        
        self.outQueue = outQueue   # Output queue (sentences)
        if inQueue is None: self.inQueue = Queue.Queue()
        else:               self.inQueue = inQueue     # Input queue (tokens)
        self.done = True

    def run(self):
        ''' Start listening for commands (tokens) from the inQueue. '''
        if debug: print 'gossypFetcher_YahooMessageBoards: running.'
        while 1:
            try:          token = self.inQueue.get()  # Get orders
            except Empty: time.sleep(1)
            if token.shutdown:
                if debug: print 'gossypFetcher_YahooMessageBoards: shutting down.'
                break
            elif token.fetch:
                self.fetchSentences()
            else:
                pass

    def html2txt(self, html, replacementString=' '):
        ''' Converts an HTML document to plain text.
            All HTML tags are replaced with replacementString. (eg. <table>)
            All HTML entities are converted to ISO-Latin 1 (eg. &amp; => &)
            Input: html (string): the source html code.
                   replacementString (string): the string which HTML tags must be replaced with.
            Output: string (the text-only version).'''
        def entity_replace(match):
            if match.group(1) == 'nbsp': return ' '
            try: return htmlentitydefs.entitydefs[match.group(1)]
            except: return ' '
            # FIXME: I should handle numeric entities too (eg. &#149;)
        text = html.replace('<br>','\n')
        text = self._tagsRemove_re.sub(replacementString,text)  # replace HTML tags
        text = self._htmlentity_re.sub(entity_replace, text)    # convert entyties (&amp;) to characters (&)
        text = text.replace(chr(151),'-')  # Hell, what's that ?  
        # FIXME: filter strange ascii codes ? (<32 and >127 ?)
        return text

    def yahoohtml2txt(self, html):
        ''' Convert a Yahoo! Message (from Yahoo! Message board) from HTML to pure text.
            Input: html (string): the html page from Yahoo Message Boards.
            Output: string : the plain text message. '''
        ''' Technical details:
        All Yahoo messages seem to start after <tr><td colspan="2"> until the next </td>
        so this is easy to extract.
        Some HTML entities have to be converted (&amp; for example), and all tags must be removed
        (like <br>)
        '''
        tag_start = '<tr><td colspan="2">'
        tag_end = '</td>'
        output = ' '
        #open(sha.new(html).hexdigest()+'.dat','w+b').write(html)
        i = html.find(tag_start)
        if i<0:
            tag_start = '<tr><td colspan=2>'  # To deal with a modified version.
            i = html.find(tag_start)
        if i>=0:
            j = html.find(tag_end,i)
            if j>=0:
                message = html[i+len(tag_start):j]  # Extract the message from HTML code.
                message = self.html2txt(message)    # Convert to pure text.
                message = message.replace('\n',' ').replace('\r',' ') # Remove newlines
                # remove spaces that are useless.
                while (message.find('  ')>-1): message = message.replace('  ',' ')
                output = message
        return output

    def getURLs(self, html):
        ''' Extracts the 3 different URLs from any Yahoo Messageboards pages:
            - URLs to sub-directories (/yahoo/)
            - URLs to topics (/bbs? ... action=topics ...)
            - URLs to messages (/bbs? ... action=m ...)
            Input: html (string): an HTML page from Yahoo! Message boards (either directory, topic or message).
            Output: a tuple (subdirs,topics,messages)
                     subdirs is a list of string. Each string is an absolute URL to a Y! Message board directory.
                     topics is a list of string. Each string is an absolute URL to a Y! Message board topic.
                     messages is a list of string. Each string is an absolute URL to a Y! Message board discussion and/or message.
                     You are warned some of these URLs may be invalid.
                     You are advised to try/except your urlopen() calls on these URLs.
        '''
        hrefs = []
        try:    hrefs = self._href_re.findall(html)
        except: pass  # Protection against 'maximum recursion limit exceeded.' error
        subdirs = []
        topics = []
        messages = []
        for href in hrefs:
            if not href.startswith('http:'):  # Relative URL --> Absolute URL.
                if not href.startswith('/'): href = '/'+href
                href = 'http://messages.yahoo.com'+href
            if href.find('messages.yahoo.com/yahoo/')>=0: # This is an url to a directory.
                subdirs.append(href)
            elif href.find('bbs?')>=0:
                if href.find('action=topic')>=0: # This is an url to a topic.
                    topics.append(href)
                elif href.find('action=m')>=0:   # This is an url to a discussion or a message.
                    messages.append(href)
        return (subdirs,topics,messages)

    def _exploreDirectories(self,num):
        ''' Explore num directories which have not been explored yet (self._dir_2do)
            and extract subdirectories and topics from them.
            Input: num (integer): number of directories to explore at once. 
            Output: none.
        '''
        i = num
        if debug and len(self._dir_2do)>0: print 'gossypFetcher_YahooMessageBoards: fetching %d directories.' % num
        while i>0 and len(self._dir_2do)>0:
            # Get a random item from directories to process, and remove it.
            dir = random.choice(self._dir_2do.keys())
            del self._dir_2do[dir]
            self._dir[dir] = 0      # Mark it explored.
            # Fetch the page:
            page = urllib.urlopen(dir).read()
            (subdirs,topics,messages) = self.getURLs(page)
            # Store subdirectories from this page which have not been encountered yet.
            for dir in subdirs:
                if not self._dir.has_key(dir):
                    self._dir_2do[dir] = 0
            # Store topics from this page which have not been encountered yet.
            for topic in topics:
                if not self._topic.has_key(topic):
                    self._topic_2do[topic] = 0
            i = i - 1

    def _exploreTopics(self,num):
        ''' Explore num topics which have not been explored yet (self._topic_2do)
            and extracts discussions from them.
            Input: num (integer): number of topics to explore at once.
            Output: none.
        '''
        i = num
        if debug and len(self._topic_2do)>0: print 'gossypFetcher_YahooMessageBoards: fetching %d topics.' % num
        while i>0 and len(self._topic_2do)>0:
            # Get a random item from topics to process, and remove it.
            topic = random.choice(self._topic_2do.keys())
            del self._topic_2do[topic]
            self._topic[topic] = 0  # Mark it explored.
            # Fetch the page:
            page = urllib.urlopen(topic).read()
            (subdirs,topics,messages) = self.getURLs(page)
            # Store discussions which have not been encountered yet.
            for disc in messages:
                if not self._disc.has_key(disc):
                    self._disc_2do[disc] = 0
            i = i - 1

    def _exploreDiscussions(self,num):
        ''' Explore num discussions which have not been explored yet (self._disc_2do)
            and extracts discussions from them.
            Input: num (integer): number of discussions to explore at once.
            Output: none.
        '''
        i = num
        if debug and len(self._disc_2do)>0: print 'gossypFetcher_YahooMessageBoards: fetching %d discussions.' % num
        while i>0 and len(self._disc_2do)>0:
            # Get a random item from discussions to process, and remove it.
            disc = random.choice(self._disc_2do.keys())
            del self._disc_2do[disc]
            self._disc[disc] = 0  # Mark it explored.
            # Fetch the page:
            page = urllib.urlopen(disc).read()
            (subdirs,topics,messages) = self.getURLs(page)
            # Store discussions which have not been encountered yet.
            for message in messages:
                self._message[message] = 0
            i = i - 1

    def _getMessages(self, num):
        ''' Choose num random messages, fetch them and extract all sentences from them.
            Input: num (integer): number of message to get sentences from.
            Output: none. New sentences are written to the output queue (self.outQueue). '''
        i = num
        sentences_count = 0
        if debug and len(self._message)>0: print 'gossypFetcher_YahooMessageBoards: fetching %d messages.' % num
        while i>0 and len(self._message)>0:
            # Get a random message URL to process, and remove it.
            message = random.choice(self._message.keys())
            del self._message[message]
            # Fetch the page:
            try:
                page = urllib.urlopen(message).read()

                # FIXME: extract plaintext message from HTML page.
                ctext = self.yahoohtml2txt(page)
                ctext = ctext.replace('...','') # Keep '...' safe.

                # Split into phrases using . ! or ? as separator.
                split1 = self._sentences_regexp.findall(ctext)

                # Put back the '...'.
                # Remove sentences that are too short or too long.
                sentences = [t.replace('','...').strip() for t in split1 if (len(t)>5 and len(t)<140)]
                sentences_count += len(sentences)
                for sentence in sentences: self.outQueue.put(sentence)
            except:
                continue  # If any problem fetching or processing the message, just skip to the next one.
            i = i - 1
        if debug and sentences_count>0: print 'gossypFetcher_YahooMessageBoards: I gave %d sentences.' % sentences_count

    def fetchSentences(self):
        ''' Fetches sentences from Usenet using Google Groups search engine. '''
        self.done = False
        if debug: print 'gossypFetcher_YahooMessageBoards: fetching.'

        # If we are starting for the first time, fetch the Yahoo Messageboards home page.
        if len(self._dir)==0:
            try:    page = urllib.urlopen('http://messages.yahoo.com/yahoo/index.html').read()
            except:
                if debug: print 'gossypFetcher_YahooMessageBoards: Could not get page from messages.yahoo.com.'
                return
            #open('page.html','w+b').write(page)
            (main_subdirs,main_topics,main_messages) = self.getURLs(page) # Get main page.
            for dir in main_subdirs:  # Store directories available on the home page.
                self._dir_2do[dir]=0

        self._getMessages(2)        # Get random messages, and extract sentences from them.
        self._exploreDirectories(2) # Explore random directories, looking for other directories and topics references.
        self._exploreTopics(2)      # Explore random topic, looking for discussions references .
        self._exploreDiscussions(2) # Explore random discussions, looking for messages references.
        self._getMessages(3)        # Get random messages, and extract sentences from them.

        # FIXME: purge those dictionnaries in order not to eat up all the memory.

        # dir=toexplore/explored topics=toexplore/explored disc.=toexplore/explored mess=toexplore.
        if debug: print 'gossypFetcher_YahooMessageBoards: dirs=%d/%d topics=%d/%d disc.=%d/%d mess.=%d' \
                        % (len(self._dir_2do),len(self._dir),len(self._topic_2do),len(self._topic),len(self._disc_2do),len(self._disc),len(self._message))

        if debug: print 'gossypFetcher_YahooMessageBoards: Done fetching.'
        self.done = True

if __name__ == "__main__":
    #os.environ['http_proxy'] = 'http://127.0.0.1:3131'
    debug = True
    out = Queue.Queue()
    f = gossypFetcher_YahooMessageBoards(out)
    f.start()
    f.inQueue.put(token.Token(fetch=1))
    f.inQueue.put(token.Token(fetch=1))
    while 1:
        try   :sentence = out.get()
        except: pass
        print sentence+'\n'
        time.sleep(3)


